#include "coreinterface.h"
#include "../include/configinterface.h"

#include <QCoreApplication>
#include <QHostAddress>
#include <QTextStream>

#include <iostream>
using namespace std;

#define CONFIG(a) _config->getConfigValue(a)
#define MKBOT_CONNECT \
	_socket->close(); \
	cout << ": Connecting to server" << endl; \
	_socket->connectToHost( \
		CONFIG("server"), \
		CONFIG("port").toInt());

QString CoreInterface::getConfigValue(QString key)
{
	return _config->getConfigValue(key);
}

QDomElement CoreInterface::getPluginConfig(QString pluginName)
{
	return _config->getPluginNode(pluginName);
}

QDomDocument CoreInterface::configDoc()
{
	return _config->getDoc();
}

void CoreInterface::joinChannel(QString channel)
{
	sendRaw(QString("JOIN %1").arg(channel));
}

void CoreInterface::partChannel(QString channel)
{
	sendRaw(QString("PART %1 :Bye!").arg(channel));
}

void CoreInterface::saveConfig()
{
	_config->saveConfig();
}

void CoreInterface::loopbackMessage(QString to, QString from, QString message)
{
	bool pm;
	to.left(1) == QString("#")
		? pm = false
		: pm = true;

	QRegExp parser;
	parser.setCaseSensitivity(Qt::CaseInsensitive);
	if ( pm )
		parser.setPattern(QString("(?:[%1])?([a-z]+)\\.([a-z]+) ?([^ ].+)?").arg(CONFIG("prefix")));
	else
		parser.setPattern(QString("[%1]([a-z]+)\\.([a-z]+) ?([^ ].+)?").arg(CONFIG("prefix")));
	if ( parser.exactMatch(message) )
	{
		if ( _pluginObjects.contains(parser.cap(1).toLower()) )
		{
			PluginInterface* plugin = _pluginObjects[parser.cap(1).toLower()];
			if ( commands.contains(plugin) )
				if ( commands[plugin].contains(parser.cap(2).toLower()) )
				{
					if ( _security->commandOk(parser.cap(1).toLower(), parser.cap(2).toLower(), from) )
						commands[plugin][parser.cap(2).toLower()]->go(to, from, pm, parser.cap(3));
					else
						reply(to, from, pm, "You do not have the neccessary permissions."); 
				}
		}
	}
					
	foreach(PluginInterface *plugin, _pluginObjects)
		if ( ! _deletedPlugins.contains(plugin) )
			if ( commands[plugin].contains("_message") )
			{
				plugin->addRef();
				commands[plugin]["_message"]->go(to, from, pm, message);
				plugin->delRef();
			}
}

bool CoreInterface::loadPlugin(QString name)
{
	if ( _pluginObjects.contains(name) )
	{
		_lastError = "Plugin already loaded.";
		return false;
	}

	PluginInterface* (*getObject)(CoreInterface*);
	QLibrary *lib = new QLibrary(QString("../plugins/%1/%1").arg(name));
	if ( ! lib->load() )
	{
		_lastError = "Could not open library";
		delete lib;
		return false;
	}
	if ( name != "config" )
		_config->loadPlugin(name);

	getObject = (PluginInterface*(*)(CoreInterface*)) lib->resolve("getObject");
	if ( getObject == 0)
	{
		_lastError = "Could not resolve symbol";
		_config->unloadPlugin(name);
		lib->unload();
		delete lib;
		return false;
	}
	_pluginLibraryHandles[name] = lib;
	QString index = name.toLower();
	_pluginObjects[index] = getObject(this);
	if ( _pluginObjects[index]->isFilter() )
		_filters.append(_pluginObjects[index]);
	
	_pluginObjects[index]->resetRef();

	if (name == "config")
		_config = (ConfigInterface*) _pluginObjects[index];
	if (name == "security")
		_security = (SecurityInterface*) _pluginObjects[index];

	cout << QString(": Loaded %1 plugin").arg(name).toStdString() << endl;
	return true;
}

bool CoreInterface::unloadPlugin(QString name, bool isReloading)
{
	if ( name == "config" && ! isReloading )
	{
		_lastError = "Config plugin can be reloaded, but not unloaded.";
		return false;
	}

	if ( name == "security" && ! isReloading )
	{
		_lastError = "Security plugin can be reloaded, but not unloaded.";
		return false;
	}

	if ( ! _pluginLibraryHandles.contains(name) )
	{
		_lastError = "Plugin is not currently loaded. Note this command is case-sensitive.";
		return false;
	}

	
	_config->unloadPlugin(name);
	

	PluginInterface *obj = _pluginObjects[name.toLower()];
	if ( _filters.contains(obj) )
		_filters.removeAt(_filters.indexOf(obj));
	unregisterPlugin(obj);
	_pluginObjects.remove(name.toLower());
	_deletedPlugins.append(obj);
	
	_toDeletePlugins.append(obj);
	_toCloseLibraries.append(_pluginLibraryHandles[name]);
	_pluginLibraryHandles.remove(name);

	cout << QString(": Unloaded %1 plugin").arg(name).toStdString() << endl;

	return true;
}

void CoreInterface::registerCommand(PluginInterface* plugin, QString command, CommandInterface* commandObject)
{
	commands[plugin][command.toLower()] = commandObject;
}

void CoreInterface::unregisterPlugin(PluginInterface* plugin)
{
	QList<CommandInterface*> removedObjects;
	foreach(CommandInterface* command, commands[plugin])
		if ( ! removedObjects.contains(command) )
		{
			delete command;
			commands[plugin].remove(commands[plugin].key(command));
			removedObjects.append(command);
		}
	commands.remove(plugin);
}

bool CoreInterface::reloadPlugin(QString name)
{
	if (unloadPlugin(name, true))
	{
		_toReloadPlugins.append(name);
		return true;
	}
	return false;
}

CoreInterface::CoreInterface() : QObject()
{
	if ( ! loadPlugin("config") )
	{
		cerr << "! Error loading config plugin: " << _lastError.toStdString() << endl;
		QCoreApplication::quit();
		return;
	}

	if ( ! loadPlugin("security") )
	{
		cerr << "! Error loading security plugin: " << _lastError.toStdString() << endl;
		QCoreApplication::quit();
		return;
	}
	_security = (SecurityInterface*) _pluginObjects["security"];
	
	QStringList pluginList = _config->getPluginList();
	foreach (QString pluginName, pluginList)
	{
		if ( pluginName == "config" || pluginName == "security" ) continue;
		if ( ! loadPlugin(pluginName) )
			cerr << QString("! Error loading %1 plugin: %2").arg(pluginName).arg(_lastError).toStdString() << endl;
	}
	_socket = new QTcpSocket();
	connect(_socket, SIGNAL(readyRead()), this, SLOT(slotReadData()));
	connect(_socket, SIGNAL(stateChanged(QAbstractSocket::SocketState)), this, SLOT(slotStateChanged(QAbstractSocket::SocketState)));
	connect(_socket, SIGNAL(error(QAbstractSocket::SocketError)), this, SLOT(slotError(QAbstractSocket::SocketError)));

	MKBOT_CONNECT
}

void CoreInterface::slotReadData()
{
	QTextStream in(_socket);
	QString data;
	do
	{
		data = in.readLine();

		foreach(PluginInterface* plugin, _pluginObjects)
			if ( commands[plugin].contains("_raw") )
			{
				plugin->addRef();
				commands[plugin]["_raw"]->go("","",false,data);
				plugin->delRef();
			}
		
		QStringList words = data.split(" ");
		
		if ( words[0].toLower() == "ping" )
		{
			sendRaw(QString("PONG %1").arg(words[1]));
		}

		if ( words.count() >= 2 )
		{
			if ( words[0].left(1) == ":" && words[1] == "001" )
				foreach(PluginInterface* plugin, _pluginObjects)
					if ( commands[plugin].contains("_connected") )
					{
						plugin->addRef();
						commands[plugin]["_connected"]->go("", "", false, "");
						plugin->delRef();
					}
	
			if ( words[0].left(1) == ":" && words[1] == "002" )
			{
				QRegExp versionParser;
				versionParser.setCaseSensitivity(Qt::CaseInsensitive);
				versionParser.setPattern("^.*running version (.*)$");
				if ( ! versionParser.exactMatch(data) )
					cerr << "Unparsable 002 line: " << data.toStdString() << endl;
				else
					_ircd = versionParser.cap(1);
			}
	
			QString type = words[1].toLower();
	
			if ( type == "privmsg" )
			{
				QRegExp parser("^:([^ !]+)![^ ]+ PRIVMSG ([^ ]+) :(.+)$", Qt::CaseInsensitive);
				if ( ! parser.exactMatch(data) )
					cerr << "! Unparsable PRIVMSG: " << data.toStdString() << endl;
				else loopbackMessage(parser.cap(2), parser.cap(1), parser.cap(3));
			}
	
			if ( type == "nick" )
			{
				QRegExp parser("^([^! ]+)![^ ]+ NICK :(.+)$", Qt::CaseInsensitive);
				if ( ! parser.exactMatch(data) )
					cerr << "! Unparsable NICK: " << data.toStdString() << endl;
				else
					foreach(PluginInterface* plugin, _pluginObjects)
						if ( commands[plugin].contains("_nickChange") )
						{
							plugin->addRef();
							commands[plugin]["_nickChange"]->go(parser.cap(1), parser.cap(2), false, "");
							plugin->delRef();
						}
			}
	
			if ( type == "quit" )
			{
				QRegExp parser("^([^! ]+)![^ ]+ QUIT :.*$", Qt::CaseInsensitive);
				if ( ! parser.exactMatch(data) )
					cerr << "! Unparsable QUIT: " << data.toStdString() << endl;
				else
					foreach(PluginInterface* plugin, _pluginObjects)
						if ( commands[plugin].contains("_userQuit") )
						{
							plugin->addRef();
							commands[plugin]["_userQuit"]->go("","" , false, parser.cap(1));
							plugin->delRef();
						}
			}
	
			if ( type == "part" )
			{
				QRegExp parser("^([^! ]+)![^ ]+ PART ([^ ]+)(?: :.*)?$", Qt::CaseInsensitive);
				if ( ! parser.exactMatch(data) )
					cerr << "! Unparsable PART: " << data.toStdString() << endl;
				else
					foreach(PluginInterface* plugin, _pluginObjects)
						if ( commands[plugin].contains("_userPart") )
						{
						plugin->addRef();
						commands[plugin]["_userPart"]->go("", "", false, parser.cap(1));
						plugin->delRef();
						}
			}
	
			if ( type == "join" )
			{
				QRegExp parser("^([^! ]+)![^ ]+ JOIN :([^ ]+)$", Qt::CaseInsensitive);
				if ( ! parser.exactMatch(data) )
					cerr << "! Unparsable JOIN: " << data.toStdString() << endl;
				else
					foreach(PluginInterface* plugin, _pluginObjects)
						if ( commands[plugin].contains("_userJoin") )
						{
							plugin->addRef();
							commands[plugin]["_userJoin"]->go(parser.cap(2), "", false, parser.cap(1));
							plugin->delRef();
						}
			}
		}
		foreach(PluginInterface* plugin, _toDeletePlugins)
			delete plugin;
		_toDeletePlugins.clear();
		_deletedPlugins.clear();
		foreach(QLibrary* lib, _toCloseLibraries)
		{
			lib->unload();
			delete lib;
		}
		_toCloseLibraries.clear();
		foreach(QString plugin, _toReloadPlugins)
			loadPlugin(plugin);
		_toReloadPlugins.clear();
	}
	while ( !in.atEnd() );
}

void CoreInterface::reply(QString to, QString from, bool pm, QString message)
{
	pm
		? sendMessage(from, message)
		: sendMessage(to, QString("%1: %2").arg(from, message));
}

void CoreInterface::sendMessage(QString to, QString message)
{
	QRegExp hostmaskParser("^([^ #!]+)!.+$");
	if ( hostmaskParser.exactMatch(to) )
		to.replace(hostmaskParser, "\1");
	sendRaw(QString("PRIVMSG %1 :%2").arg(to).arg(message));
}

void CoreInterface::sendNotice(QString to, QString message)
{
	QRegExp hostmaskParser("^([^ #!]+)!.+$");
	if ( hostmaskParser.exactMatch(to) )
		to.replace(hostmaskParser, "\1");
	sendRaw(QString("NOTICE %1 :%2").arg(to).arg(message));
}

void CoreInterface::sendRaw(QString command)
{
	foreach(PluginInterface* plugin, _filters)
	{
		plugin->addRef();
		command = plugin->filterOutput(command);
		plugin->delRef();
	}

	if ( command == "" ) return;
	command += "\n";
	_socket->write(command.toLatin1());
}

void CoreInterface::sendAction(QString to, QString message)
{
	sendRaw(QString("PRIVMSG %1 :\001ACTION %2\001").arg(to).arg(message));
}

void CoreInterface::slotError(QAbstractSocket::SocketError socketError)
{
	switch ( socketError )
	{
		case QAbstractSocket::ConnectionRefusedError:
			cerr << "! Connection refused" << endl;
			QCoreApplication::quit();
			break;
		case QAbstractSocket::RemoteHostClosedError:
			cerr << "! Remote host closed the connection" << endl;
			QCoreApplication::quit();
			break;
		case QAbstractSocket::HostNotFoundError:
			cerr << "! Host not found" << endl;
			QCoreApplication::quit();
			break;
		case QAbstractSocket::NetworkError:
			cerr << "! An unknown network error occured" << endl;
			QCoreApplication::quit();
			break;
		case QAbstractSocket::SocketTimeoutError:
			cerr << "! Socket timeout" << endl;
			QCoreApplication::quit();
			break;
	}
}

void CoreInterface::slotStateChanged(QAbstractSocket::SocketState newState)
{
	if ( newState == QAbstractSocket::ConnectedState )
	{	
		QString address = _socket->localAddress().toString();
		if ( address == "" ) address = "127.0.0.1";
		
		cout << ": Connected to server" << endl;
		sendRaw(QString("nick %1").arg(CONFIG("nick")));
		sendRaw(QString("user MKbot %1 %2 MKbot\n").arg(
			address).arg(
			CONFIG("server")));
	}
}

#include "moc_coreinterface.cpp"
