#include "../quotescommands.h"
#include "../quote.h"

#include <QDateTime>

namespace QuotesCommands
{
	add::add(CoreInterface* core, QHash<unsigned int, Quote*>* quotes, Quote** lastQuote, QHash<QString, QStringList>* log, unsigned int* highestQuote)
	{
			_core = core;
			_quotes = quotes;
			_lastQuote = lastQuote;
			_log = log;
			_highestQuote = highestQuote;
	}

	QString add::description()
	{
		return "Adds a quote to the database";
	}

	QStringList add::syntax()
	{
		return (QStringList()
			<< ""
			<< "<nick>"
			<< "/<pattern>/"
			<< "<from> <length>"
			<< "/<pattern>/ /<patternTo>/");
	}
	
	QStringList add::helpText()
	{
		return QStringList()
			<< "Adds a quote to the database. If no parameters are given, the last line is quoted."
			<< "If a nickname is specified, the last line that that person said is quoted."
			<< "If a single pattern is specified, the last line matching 'pattern' is quoted."
			<< "If from and length are specified, the block of text starting from 'from' lines previously and lasting 'length' lines is quoted."
			<< "If two patterns are specified, multiple lines are quoted, starting with the last line matching the first pattern, ending with the last line matching the second pattern."
			<< "For exact pattern syntax, see http://doc.trolltech.com/4.0/qregexp.html";
	}
	
	void add::go(QString to, QString from, bool pm, QString params)
	{
		QRegExp parser;
		parser.setCaseSensitivity(Qt::CaseInsensitive);
		
		if ( pm )
		{
			_core->sendMessage(from, "Can not quote PMs");
			return;
		}
		
		QStringList words = params.split(" ");
		
		if ( params == "" ) // Quote the previous line
		{
			if ( (*_log)[to.toLower()].count() == 0 )
			{
				_core->reply(to, from, pm, "No lines in log");
				return;
			}
			unsigned int quoteNum = ++(*_highestQuote);
			Quote* newQuote = new Quote();
			newQuote->setQuotedBy(from.toLower());
			newQuote->setTime(QDateTime::currentDateTime().toString());
			newQuote->addLine((*_log)[to.toLower()][0]);
			
			parser.setPattern("^<([^>]+)>.*$");
			parser.exactMatch((*_log)[to.toLower()][0]);
			newQuote->addQuotedNick(parser.cap(1).toLower());
			
			(*_quotes)[quoteNum] = newQuote;
			*_lastQuote = newQuote;
			saveQuote(quoteNum, _core, _quotes, _highestQuote);
			
			parser.setPattern("^(.{0,40}).*$");
			parser.exactMatch((*_log)[to.toLower()][0]);
			_core->reply(to, from, pm, QString("Added quote #%1: %2...").arg(quoteNum).arg(parser.cap(1)));
			return;
		}
	
		parser.setPattern("^/([^/]+)/$");

		if ( parser.exactMatch(params) ) // match by regex
		{
			parser.setPattern(parser.cap(1));
			if ( ! parser.isValid() )
			{
				_core->reply(to, from, pm, QString("Invalid pattern: %1").arg(parser.errorString()));
				return;
			}
			QStringList log = (*_log)[to.toLower()];
			for ( int i = 0; i < log.count(); i++ )
				if ( parser.indexIn(log[i]) != -1 )
				{
					parser.setPattern("^<([^>]+)>.*$");
					parser.exactMatch(log[i]);
					unsigned int quoteNum = ++(*_highestQuote);
					Quote* newQuote = new Quote();
					newQuote->setQuotedBy(from.toLower());
					newQuote->setTime(QDateTime::currentDateTime().toString());
					newQuote->addLine(log[i]);
					newQuote->addQuotedNick(parser.cap(1));

					(*_quotes)[quoteNum] = newQuote;
					*_lastQuote = newQuote;
					saveQuote(quoteNum, _core, _quotes, _highestQuote);

					parser.setPattern("^(.{0,40}).*$");
					parser.exactMatch(log[i]);
					_core->reply(to, from, pm, QString("Added quote #%1: %2...").arg(quoteNum).arg(parser.cap(1)));
				}
			_core->reply(to, from, pm, "No quote found.");
			return;
		}
	
		if ( words.count() == 1 ) // Quote nick specified
		{
			parser.setPattern(QString("^<(%1)>.*$").arg(words[0]));
			for ( int i = 0; i < (*_log)[to.toLower()].count(); i++ )
				if ( parser.exactMatch((*_log)[to.toLower()][i]) )
				{
					unsigned int quoteNum = ++(*_highestQuote);
					Quote* newQuote = new Quote();
					newQuote->setQuotedBy(from.toLower());
					newQuote->setTime(QDateTime::currentDateTime().toString());
					newQuote->addLine((*_log)[to.toLower()][i]);
					newQuote->addQuotedNick(parser.cap(1));

					(*_quotes)[quoteNum] = newQuote;
					*_lastQuote = newQuote;
					saveQuote(quoteNum, _core, _quotes, _highestQuote);

					parser.setPattern("^(.{0,40}).*$");
					parser.exactMatch((*_log)[to.toLower()][i]);
					_core->reply(to, from, pm, QString("Added quote #%1: %2...").arg(quoteNum).arg(parser.cap(1)));
					return;
				}
			_core->reply(to, from, pm, QString("No quote found for nick '%1'.").arg(words[0]) );
			return;
		}

		parser.setPattern("^/([^/]+)/ /([^/]+)/$");
		if ( parser.exactMatch(params) )
		{
			QRegExp startRx(parser.cap(1), Qt::CaseInsensitive);
			QRegExp endRx(parser.cap(2), Qt::CaseInsensitive);
			int startPos = -1;
			int endPos = -1;
			
			QStringList log = (*_log)[to.toLower()];
			for ( int i = 0; i < log.count(); i++ )
			{
				if ( startPos == -1 && startRx.indexIn(log[i]) != -1 ) startPos = i;
				if ( endPos == -1 && endRx.indexIn(log[i]) != -1 ) endPos = i;
				if ( startPos != -1 && endPos != -1 )
					break;
			}
			
			if ( startPos == -1 || endPos == -1 )
			{
				_core->reply(to, from, pm, "Quote not found");
				return;
			}

			if ( startPos < endPos )
			{
				_core->reply(to, from, pm, "Quote end before quote start.");
				return;
			}

			QStringList lines = log.mid(endPos, (startPos - endPos) + 1);
			
			Quote* newQuote = new Quote();
			newQuote->setQuotedBy(from);
			newQuote->setTime(QDateTime::currentDateTime().toString());
			for ( int i = lines.count() - 1; i >= 0; i-- )
			{
				newQuote->addLine(lines[i]);
				parser.setPattern("^<([^>]+)>.*$");
				parser.exactMatch(lines[i]);
				if ( ! newQuote->quoted().contains(parser.cap(1).toLower()) )
					newQuote->addQuotedNick(parser.cap(1).toLower());
			}

			unsigned int quoteNum = ++(*_highestQuote);

			(*_quotes)[quoteNum] = newQuote;
			*_lastQuote = newQuote;
			saveQuote(quoteNum, _core, _quotes, _highestQuote);

      parser.setPattern("^(.{0,30}).*$");
      parser.exactMatch(newQuote->quoteLines()[0]);
      QString startPreview = parser.cap(1);
      parser.setPattern("^.*(.{0,30})$");
      parser.exactMatch(newQuote->quoteLines().last());
      QString endPreview = parser.cap(1);

      _core->reply(to, from, pm, QString("Added quote #%1: %2...%3").arg(quoteNum).arg(startPreview).arg(endPreview));
      return;
		}

		if ( words.count() == 2 ) // <line from> <length>
		{
			bool startInt; words[0].toInt(&startInt);
			bool endInt; words[1].toInt(&endInt);

			if ( startInt && words[0].toInt() > (*_log)[to.toLower()].count() )
			{
				_core->reply(to, from, pm, "Log not long enough");
				return;
			}
			if ( startInt && endInt && words[0].toInt() < words[1].toInt() )
			{
				_core->reply(to, from, pm, "Length longer than number of lines since start of quote");
				return;
			}

			if ( ! (startInt && endInt) )
			{
				_core->reply(to, from, pm, "Incorrect syntax. See Help.Help Quotes.Add");
				return;
			}
			
			Quote* newQuote = new Quote();
			newQuote->setQuotedBy(from);
			newQuote->setTime(QDateTime::currentDateTime().toString());
			for ( int i = words[0].toInt() - 1; i > (words[0].toInt() - words[1].toInt()) - 1; i-- )	
			{
				newQuote->addLine((*_log)[to.toLower()][i]);
				parser.setPattern("^<([^>]+)>.*$");
				parser.exactMatch((*_log)[to.toLower()][i]);
				if ( ! newQuote->quoted().contains(parser.cap(1).toLower()) )
					newQuote->addQuotedNick(parser.cap(1).toLower());
			}
			
			unsigned int quoteNum = ++(*_highestQuote);

			(*_quotes)[quoteNum] = newQuote;
			*_lastQuote = newQuote;
			saveQuote(quoteNum, _core, _quotes, _highestQuote);

			parser.setPattern("^(.{0,30}).*$");
			parser.exactMatch(newQuote->quoteLines()[0]);
			QString startPreview = parser.cap(1);
			parser.setPattern("^.*(.{0,30})$");
			parser.exactMatch(newQuote->quoteLines().last());
			QString endPreview = parser.cap(1);

			_core->reply(to, from, pm, QString("Added quote #%1: %2...%3").arg(quoteNum).arg(startPreview).arg(endPreview));
			return;
		}
		_core->reply(to, from, pm, "Incorrect syntax. See Help.Help Quotes.Add");
	}
};
