#include "../quotescommands.h"
#include "../quote.h"

namespace QuotesCommands
{
	get::get(CoreInterface* core, QHash<unsigned int, Quote*>* quotes, Quote** lastQuote)
	{
			_core = core;
			_quotes = quotes;
			_lastQuote = lastQuote;
	}

	QString get::description()
	{
		return "Retrieves a quote from the database";
	}

	QStringList get::syntax()
	{
		return (QStringList()
			<< ""
			<< "<number>"
			<< "<nick>"
			<< "/<pattern>/");
	}
	
	QStringList get::helpText()
	{
		return QStringList("Retrieves a quote from the database. If no parameters are given, the last quote is retrieved. If a number is specified the quote with that number is retrieved. If a nickname is specified, a random quote of that person is retrieved. If a pattern is specified, a random quote matching that pattern is retrieved. For exact syntax of patterns, see http://doc.trolltech.com/4.0/qregexp.hgtml");
	}
	
	void get::go(QString to, QString from, bool pm, QString params)
	{
		QRegExp parser;
		parser.setCaseSensitivity(Qt::CaseInsensitive);
		// See if this is a lastquote
		if ( params == "" )
		{
			LASTQUOTE_CHECK;
			outputQuote(to, from, pm, _quotes->key(*_lastQuote), _quotes, _core);
			return;
		}

		// See if this is quote-by-number
		parser.setPattern("[0-9]+");
		if ( parser.exactMatch(params) )
		{
			unsigned int quoteNum = params.toUInt();
			if ( ! _quotes->contains(quoteNum) )
			{
				_core->reply(to, from, pm, "No such quote.");
				return;
			}

			outputQuote(to, from, pm, quoteNum, _quotes, _core);
			*_lastQuote = (*_quotes)[quoteNum];
			return;
		}

		// See if this is quote-by-nick
		parser.setPattern("[^/].*");
		if ( parser.exactMatch(params) )
		{
			QList<Quote*> nickQuotes;
			foreach(Quote* value, _quotes->values())
				if ( value->quoted().contains(params.toLower()) )
					nickQuotes.append(value);
					
			if ( nickQuotes.count() == 0 )
			{
				_core->reply(to, from, pm, "No quotes found.");
				return;
			}
			// Get a random number, convert it to a float between 0 and 1, times by count and truncate
			//  to int to get 0 <= quoteNum < nickQuotes.count(), then get that member of nickQuotes[]
			Quote* quote = nickQuotes[(unsigned int) ((double)nickQuotes.count()*(float)rand()/(RAND_MAX+1.0))];
			outputQuote(to, from, pm, _quotes->key(quote), _quotes, _core);
			*_lastQuote = quote;
			return;
		}

		// See if this is quote-by-regex
		parser.setPattern("/(.*)/");
		if ( parser.exactMatch(params) )
		{
			parser.setPattern(parser.cap(1));
		
			QList<Quote*> matchingQuotes;
			foreach(Quote* value, _quotes->values())
				if ( parser.indexIn(value->quoteLines().join("\n")) != -1 )
					matchingQuotes.append(value);

			if ( matchingQuotes.count() == 0 )
			{
				_core->reply(to, from, pm, "No quotes found.");
				return;
			}

			// See quote-by-nick for explanation of this
			Quote* quote = matchingQuotes[(unsigned int) ((double)matchingQuotes.count()*(float)rand()/(RAND_MAX+1.0))];
			outputQuote(to, from, pm, _quotes->key(quote), _quotes, _core);
			*_lastQuote = quote;
			return;
		}

		_core->reply(to, from, pm, "Invalid syntax, see Help.Help Quotes.Get");
	}
};
