#include "security.h"

#include <iostream>
using namespace std;

#include <QRegExp>
#include <QStringList>

extern "C" PluginInterface* getObject(CoreInterface *core)
{
	return (PluginInterface*) new Security(core);
}

QString Security::title()
{
	return "Security";
}

QString Security::description()
{
	return "Checks that users have permission to execute commands";
}

Security::Security(CoreInterface* core) : SecurityInterface()
{
	_core = core;
	_rulesTable = new QList<QStringList>;

	// Register commands
	_core->registerCommand(this, "auth", new SecurityCommands::auth(_core));
	_core->registerCommand(this, "list", new SecurityCommands::list(_core));
	_core->registerCommand(this, "grant", new SecurityCommands::grant(_core, _rulesTable));
	_core->registerCommand(this, "revoke", new SecurityCommands::revoke(_core, _rulesTable));
	_core->registerCommand(this, "_raw", new SecurityCommands::_raw(_core, &_authedNicks));

	SecurityCommands::_userQuit* userQuit = new SecurityCommands::_userQuit(_core, &_authedNicks);
	_core->registerCommand(this, "_userQuit", userQuit);
	_core->registerCommand(this, "_nickChange", userQuit);
	_core->registerCommand(this, "_userPart", userQuit);

	QDomElement root = _core->getPluginConfig("security");

	// Check if we have existing rules
	for ( QDomNode i = root.firstChild(); !i.isNull(); i = i.nextSibling() )
		if ( i.nodeName() == "rule" )
		{
			loadRules();
			return;
		}
	
	// Add a default allow rule
	GRANT("//", "//", "//");
}

bool Security::isAuthed(QString nick)
{
	return _authedNicks.contains(nick.toLower());
}

QStringList Security::helpText()
{
	return (QStringList()
		<< "This plugin makes sure that users have permission to execute any"
		<< "commands that they try to. Note that any user can execute"
		<< "'Security.auth' to identify themselves to the bot, irrespective of permissions.");
}

bool Security::commandOk(QString plugin, QString command, QString from)
{
	bool foundWithAuth = false;
	if ( from == "@@@CORE@@@" ) return true;
	if ( plugin.toLower() == "security" && command.toLower() == "auth" ) return true;

	foreach(QStringList rule, *_rulesTable)
	{
		QString ruleUser = rule[0];
		QString rulePlugin = rule[1];
		QString ruleCommand = rule[2];

		QRegExp parser;
		parser.setCaseSensitivity(Qt::CaseInsensitive);
		parser.setPattern("^/(.*)/$");

		QRegExp subParser;
		subParser.setCaseSensitivity(Qt::CaseInsensitive);

		if ( parser.exactMatch(ruleUser) )
		{
			subParser.setPattern(parser.cap(1));
			if ( subParser.indexIn(from) == -1 ) continue;
		}
		else if ( ruleUser.toLower() != from.toLower() && ruleUser != "*" ) continue;

    if ( parser.exactMatch(rulePlugin) )
    {
      subParser.setPattern(parser.cap(1));
      if ( subParser.indexIn(plugin) == -1 ) continue;
    }
    else if ( rulePlugin.toLower() != plugin.toLower() ) continue;

    if ( parser.exactMatch(ruleCommand) )
    {
      subParser.setPattern(parser.cap(1));
      if ( subParser.indexIn(command) == -1 ) continue;
    }
    else if ( ruleCommand.toLower() != command.toLower() ) continue;

    if ( ruleUser != "*" && ! _authedNicks.contains(from.toLower()) )
    {
      foundWithAuth = true;
      continue;
    }

    return true;
	}

	if ( foundWithAuth ) _core->sendMessage(from, QString("Please identify with nickserv and type '/msg %1 security.auth'").arg(_core->getConfigValue("nick")));
	return false;
}

void Security::loadRules()
{
	QDomElement root = _core->getPluginConfig("security");
	for ( QDomNode i = root.firstChild(); !i.isNull(); i = i.nextSibling() )
	{
		if ( i.nodeName() != "rule" ) continue;

		QString ruleUser = "", rulePlugin = "", ruleCommand = "";
		for ( QDomNode j = i.firstChild() ; !j.isNull(); j = j.nextSibling() )
		{
			if ( j.nodeName() == "user" ) ruleUser = j.firstChild().nodeValue();
			if ( j.nodeName() == "plugin" ) rulePlugin = j.firstChild().nodeValue();
			if ( j.nodeName() == "command" ) ruleCommand = j.firstChild().nodeValue();
		}
		_rulesTable->append(QStringList() << ruleUser << rulePlugin << ruleCommand);
	}
}
